"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var built_ins_1 = require("./built-ins");
var global_state_1 = require("./global-state");
var reaction_1 = require("./reaction");
var utils_1 = require("./utils");
var MAX_RUN_COUNT = 1000;
/**
 * Create observable wrapper.
 */
function createObservableObject(obj) {
    if (obj === void 0) { obj = {}; }
    if (utils_1.isPrimitive(obj)) {
        throw TypeError("dob not support " + obj + ", because it is a basic type.");
    }
    if (global_state_1.globalState.proxies.has(obj)) {
        return global_state_1.globalState.proxies.get(obj);
    }
    // Proxy inert packaging.
    return toObservable(obj);
}
/**
 * Create observable decorator.
 */
function createObservableObjectDecorator(target) {
    function wrap() {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        return createObservableObject(new (target.bind.apply(target, [void 0].concat(args)))());
    }
    return wrap;
}
/**
 * Create observable object from origin object.
 */
function toObservable(obj) {
    if (Object.getOwnPropertySymbols(obj).indexOf(global_state_1.globalState.ignoreDynamicSymbol) > -1) {
        // Ignore observable
        return obj;
    }
    var dynamicObject;
    var builtIn = built_ins_1.default.get(obj.constructor);
    if (typeof builtIn === "function" || typeof builtIn === "object") {
        // support: map weakMap set weakSet.
        dynamicObject = builtIn(obj, bindCurrentReaction, queueRunReactions, getProxyValue);
    }
    else if (!builtIn) {
        dynamicObject = new Proxy(obj, {
            get: function (target, key, receiver) {
                var value = Reflect.get(target, key, receiver);
                // Get origin object.
                if (key === "$raw") {
                    return target;
                }
                // Getter event.
                global_state_1.globalState.event.emit("get", { target: target, key: key, value: value });
                bindCurrentReaction(target, key);
                value = getProxyValue(target, key, value);
                return value;
            },
            set: function (target, key, value, receiver) {
                var oldValue = Reflect.get(target, key, receiver);
                // If the new value is an object, the original object is preferentially taken.
                if (typeof value === "object" && value) {
                    value = value.$raw || value;
                }
                global_state_1.globalState.event.emit("set", { target: target, key: key, value: value, oldValue: oldValue });
                var result = Reflect.set(target, key, value, receiver);
                // If the length of attribute changed, or the new value is different from the old value, trigger observer queue task.
                // This step should be done after `Reflect.set` to ensure that the trigger is using a new value.
                if (key === "length" || value !== oldValue) {
                    queueRunReactions(target, key);
                }
                return result;
            },
            deleteProperty: function (target, key) {
                var hasKey = Reflect.has(target, key);
                global_state_1.globalState.event.emit("deleteProperty", { target: target, key: key });
                var result = Reflect.deleteProperty(target, key);
                if (hasKey) {
                    queueRunReactions(target, key);
                }
                return result;
            }
        });
    }
    else {
        dynamicObject = obj;
    }
    global_state_1.globalState.proxies.set(obj, dynamicObject);
    global_state_1.globalState.proxies.set(dynamicObject, dynamicObject);
    global_state_1.globalState.originObjects.set(dynamicObject, obj);
    global_state_1.globalState.objectReactionBindings.set(obj, new Map());
    return dynamicObject;
}
/**
 * Ensure proxy value
 */
function getProxyValue(target, key, value) {
    // If the value is a HTMLElement object, it returns directly to the original object, because the original object cannot be encapsulated.
    if (typeof window !== "undefined" && value instanceof HTMLElement) {
        return value;
    }
    // If the value is an object, the proxy object is taken priority.
    var resultIsObject = typeof value === "object" && value;
    var existProxy = resultIsObject && global_state_1.globalState.proxies.get(value);
    if (resultIsObject) {
        return existProxy || toObservable(value);
    }
    return value;
}
/**
 * Run reactions for object's propertyKey.
 * If has running queue, reactions will be merged in it.
 */
function queueRunReactions(target, key) {
    // If in strict mode, and not in the batch, throw error.
    if (global_state_1.globalState.strictMode && global_state_1.globalState.batchDeep === 0) {
        throw Error("You are not allowed to modify observable value out of Action.");
    }
    var keyBinder = utils_1.getBinder(target, key).keyBinder;
    Array.from(keyBinder).forEach(function (reaction) {
        if (utils_1.inAction()) {
            global_state_1.globalState.pendingReactions.add(reaction);
        }
        else {
            // Not in Action, added to the queue if the queue already has a value, or directly execute if the queue has no value.
            if (global_state_1.globalState.pendingReactions.size === 0) {
                runReaction(reaction);
            }
            else {
                global_state_1.globalState.pendingReactions.add(reaction);
                runPendingReactions();
            }
        }
    });
}
function runReaction(reaction) {
    reaction.run();
}
exports.runReaction = runReaction;
/**
 * Execute pending reaction queue.
 * If in action, continue adding to actionQueuedObservers.
 * If not in the action, add a direct implementation.
 */
function runPendingReactions() {
    // The number of queue executions.
    var currentRunCount = 0;
    global_state_1.globalState.pendingReactions.forEach(function (observer) {
        currentRunCount++;
        if (currentRunCount >= MAX_RUN_COUNT) {
            global_state_1.globalState.pendingReactions.clear();
            return;
        }
        runReaction(observer);
    });
    // Clear pending reactions.
    global_state_1.globalState.pendingReactions.clear();
}
/**
 * Bind the current reaction
 */
function bindCurrentReaction(object, key) {
    // Add listener to this key, which not be tracked in the runInAction
    // InBatch as long as non zero, that runInAction is over
    if (!global_state_1.globalState.currentReaction || utils_1.inAction()) {
        return;
    }
    var keyBinder = utils_1.getBinder(object, key).keyBinder;
    // If this key is not bound with the current reaction, bind it.
    if (!keyBinder.has(global_state_1.globalState.currentReaction)) {
        keyBinder.add(global_state_1.globalState.currentReaction);
        global_state_1.globalState.currentReaction.addBinder(keyBinder);
    }
}
function isObservable(obj) {
    return (global_state_1.globalState.proxies.get(obj) === obj);
}
exports.isObservable = isObservable;
/**
 * Self run reactions
 */
function observe(callback, delay) {
    var reaction = new reaction_1.Reaction("observe", function () {
        reaction.track(callback);
    }, delay);
    // Run in initialization
    if (utils_1.inAction()) {
        //  If in the action, directly added to the queue, such as the implementation of the action will automatically execute this queue.
        global_state_1.globalState.pendingReactions.add(reaction);
    }
    else {
        reaction.run();
    }
    return {
        unobserve: function () {
            reaction.dispose();
        }
    };
}
exports.observe = observe;
/**
 * Enter next batch.
 */
function startBatch() {
    if (global_state_1.globalState.batchDeep === 0) {
        // If starting a new queue from deep 0, clear the original queue.
        global_state_1.globalState.pendingReactions = new Set();
    }
    global_state_1.globalState.batchDeep++;
    global_state_1.globalState.event.emit("startBatch", null);
}
exports.startBatch = startBatch;
/**
 * Exit the current batch.
 */
function endBatch() {
    if (--global_state_1.globalState.batchDeep === 0) {
        runPendingReactions();
    }
    global_state_1.globalState.event.emit("endBatch", null);
}
exports.endBatch = endBatch;
function Action(arg1, arg2, arg3) {
    if (arg2 === undefined) {
        return runInAction.call(null, arg1, arg1.name);
    }
    return actionDecorator.call(null, arg1, arg1.constructor.name + "." + arg2, arg3);
}
exports.Action = Action;
function actionDecorator(target, propertyKey, descriptor) {
    var func = descriptor.value;
    return {
        get: function () {
            var _this = this;
            return function () {
                var args = [];
                for (var _i = 0; _i < arguments.length; _i++) {
                    args[_i] = arguments[_i];
                }
                return runInAction(func.bind.apply(func, [_this].concat(args)), propertyKey);
            };
        }
    };
}
function runInAction(fn, debugName) {
    global_state_1.globalState.event.emit("runInAction", debugName);
    startBatch();
    try {
        return fn();
    }
    finally {
        endBatch();
    }
}
// ==============================================
// Action [end]
// ==============================================
function observable(target) {
    if (target === void 0) { target = {}; }
    if (typeof target === "function") {
        return createObservableObjectDecorator(target);
    }
    else {
        return createObservableObject(target);
    }
}
exports.observable = observable;
/**
 * The object will not listen, when use this decorator.
 */
function Static(obj) {
    Object.defineProperty(obj, global_state_1.globalState.ignoreDynamicSymbol, {
        value: true
    });
    return obj;
}
exports.Static = Static;
//# sourceMappingURL=observer.js.map